@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2000-2005  The PARI group.
+Copyright (C) 2000-2013  The PARI group.
 
 This file is part of the GP2C package.
 
@@ -10,7 +10,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
 
 #include "config.h"
 #include <stdlib.h>
@@ -19,16 +19,36 @@ Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
 #include "header.h"
 #include "patchfunc.h"
 extern int indent;
+
+int has_iter(char *s)
+{
+  int nf=getfunc(s);
+  return !!lfunc[nf].iter;
+}
+
+void patchfunction(struct patchfunc_s *p)
+{
+  int nf=getfunc(p->gpname);
+  lfunc[nf].proto.code=p->code;
+  functype(lfunc[nf])=p->type;
+  funcmode(lfunc[nf])=p->mode;
+  lfunc[nf].spec=p->spec;
+}
+
 void patchfunclist(void)
 {
   int i;
   for(i=0;patchfunc[i].gpname;i++)
+    patchfunction(patchfunc+i);
+  if (!has_iter("forvec"))
   {
-    int nf=getfunc(patchfunc[i].gpname);
-    lfunc[nf].proto.code=patchfunc[i].code;
-    functype(lfunc[nf])=patchfunc[i].type;
-    funcmode(lfunc[nf])=patchfunc[i].mode;
-    lfunc[nf].spec=patchfunc[i].spec;
+    struct patchfunc_s pforvec = {"forvec",NULL,Gnotype,(1<<Msemicolon),GPforvec};
+    patchfunction(&pforvec);
+  }
+  if (!has_iter("forprime"))
+  {
+    struct patchfunc_s pforprime = {"forprime",NULL,Gnotype,(1<<Msemicolon),GPforprime};
+    patchfunction(&pforprime);
   }
 }
 void printlistfunc(FILE *fout)
@@ -87,18 +107,18 @@ int guesstype(int n)
     if (!strcmp(entryname(n),"length"))
       return Glg;
   default: /*FALL THROUGH*/
-    return Gnotype;
+    return Ggen;
   }
 }
 
 void genequal(int aff, const char *func, int *var, int *binf, int *t)
 {
-  if (tree[aff].f!=Faffect)
+  if (tree[aff].f!=Fassign)
     die(aff,"incorrect syntax for %s",func);
   checkisvar(tree[aff].x,func,var,t);
   *binf=tree[aff].y;
 }
-void genequaltosmall(int binf,int bsup, int *tv)
+void genequaltosmall(int binf,int bsup, int *tv, int def)
 {
   if (*tv==Gnotype)
   {
@@ -107,11 +127,10 @@ void genequaltosmall(int binf,int bsup, int *tv)
     if  (is_subtype(tinf,Gsmall) && is_subtype(tsup,Gsmall))
       *tv=Gsmall;
     else
-      *tv=Ggen;
+      *tv=def;
   }
 }
 
-int addseqright(int seq, int n);
 int gengerepile(int bl)
 {
   int arg[4];
@@ -122,7 +141,7 @@ int gengerepile(int bl)
   pushvar(arg[1],0,Gpari_sp,
       newcall("_stack_lim",newnode(Flistarg,arg[0],newsmall(1))));
   arg[2]=newsmall(bl);
-  return listtoseq(arg,3);
+  return stacktoargs(arg,3);
 }
 
 int newgetg(int v, int l, const char *t)
@@ -131,31 +150,183 @@ int newgetg(int v, int l, const char *t)
           v, l), newstringnode(t,-1)));
 }
 
-void makesubblock_nolabel(int n)
+void makesubblock_ctx(int n, int savx)
 {
-  int savx=s_ctx.n;
-  int bl3=newblock();
-  genblock_nolabel(n,-1);
-  makeblock(bl3,n,newleaf(n),-1,savx);
-}
-
-void makesubblock(int n)
-{
-  int savx=s_ctx.n;
   int bl3=newblock();
   genblock(n,-1);
   makeblock(bl3,n,newleaf(n),-1,savx);
 }
 
-Clabel_t makesubblock_label(int n)
+void makesubblock(int n)
+{
+  makesubblock_ctx(n, s_ctx.n);
+}
+
+Clabel_t genblock_label(int n, int p, int type)
+{
+  Clabel_t t;
+  static int labie=0;
+  int l=newlabel(type);
+  if (type==Liferr)
+    label[l].ie=labie++;
+  genblock(n,p);
+  if (type==Liferr)
+    labie--;
+  t=label[l];
+  stack_pop_safe(&s_label,l);
+  return t;
+}
+
+Clabel_t makesubblock_label(int n, int t)
 {
   int savx=s_ctx.n;
   int bl3=newblock();
-  Clabel_t lab=genblock_label(n,-1);
+  Clabel_t lab=genblock_label(n,-1,t);
   makeblock(bl3,n,newleaf(n),-1,savx);
   return lab;
 }
 
+int check_labels(int k, int t)
+{
+  int l=s_label.n;
+  while(k--)
+  {
+    if (!l) return 2;
+    if (label[--l].type==t)
+      return 1;
+  }
+  return 0;
+}
+
+int get_iferr(int k)
+{
+  int i, l=s_label.n;
+  for (i=s_label.n-k; i<l; i++)
+  {
+    if (label[i].type==Liferr)
+      return label[i].ie;
+  }
+  return 0;
+}
+
+int
+geniterproto(int n, long fl, gpfunc *gp, int *argo, int *seq, int *var, int *tv)
+{
+  int arg[STACKSZ];
+  const char *p=gp->proto.code;
+  char c;
+  PPproto mod;
+  int nb=genlistargs(n,arg,0,STACKSZ);
+  int i=0, k=0, binf;
+  while((mod=parseproto(&p,&c))!=PPend)
+  {
+    if (i<nb && tree[arg[i]].f!=Fnoarg && (mod==PPdefault || mod==PPdefaultmulti))
+      mod=PPstd;
+    switch(mod)
+    {
+    case PPstd:
+      if (i>=nb)
+        die(n,"too few arguments");
+      switch(c)
+      {
+      case '=':
+        genequal(arg[i], gp->gpname, var, &binf, tv);
+        if (fl)
+        {
+          tree[*var].t = *tv;
+          argo[k++] = *var;
+          tree[binf].t = guesstype(binf);
+        }
+        argo[k++] = binf;
+        break;
+      case 'V':
+        checkisvar(arg[i], gp->gpname, var, tv);
+        if (fl)
+        {
+          tree[*var].t = *tv;
+          argo[k++] = *var;
+        }
+        break;
+      case 'I':
+        *seq = arg[i];
+        break;
+      default:
+        if (fl)
+          tree[arg[i]].t = guesstype(arg[i]);
+        argo[k++] = arg[i];
+      }
+      break;
+    case PPdefault:
+    case PPdefaultmulti:
+      if (i<nb) tree[arg[i]].t = guesstype(arg[i]);
+      argo[k++] = i<nb ? arg[i] : GNOARG;
+      break;
+    case PPstar:
+      i=nb-1;
+      break;
+    case PPauto: /* Skip */
+      i--;
+      break;
+    default:
+      break;
+    }
+    i++;
+  }
+  return k;
+}
+
+void
+genblockfunciter(int n, gpfunc *gp)
+{
+  int arg[STACKSZ];
+  int var=-1, tv=Gnotype;
+  int bseq=-1,aseq=-1,ret=-1,vprime,seq,cseq;
+  int gpm=-1,bl0,bl1,bl2;
+  int sav0, savx;
+  Clabel_t lab;
+  gpiterator * iter;
+  const char * init, *next;
+  int nb = geniterproto(n, 1, gp, arg, &seq, &var, &tv);
+  gpdescarg * da = descfindrulesdsc(nb, arg, gp->iter);
+  if (!da) die(n, "no suitable iterator found");
+  iter = &(da->iter);
+  init = lfunc[iter->init].gpname;
+  next = lfunc[iter->next].gpname;
+  sav0=s_ctx.n;
+  bl0=newblock();
+  vprime=newnode(Fentry,newentry("iter"),-1);
+  pushvar(vprime,1<<Cuser,iter->type,-1);
+  nb = geniterproto(n, 0, gp, arg+1, &seq, &var, &tv)+1;
+  arg[0] = vprime;
+  bseq = newcall(init, stacktoargs(arg,nb));
+  genblock(bseq,-1);
+  savx=s_ctx.n;
+  bl1=newblock();
+  bl2=newblock();
+  if (autogc)
+    gpm=gengerepile(bl2);
+  if (tv==Gnotype)
+  {
+    tree[vprime].t = iter->type;
+    gpdescarg *da=descfindrules(1, &vprime, lfunc+iter->next);
+    tv = da->type;
+  }
+  pushvar(var,1<<Cuser|1<<Ccompo,tv,-1);
+  lab=makesubblock_label(seq,Lbreak);
+  if (lab.ne)
+    seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
+  if (autogc)
+    seq=addseqright(seq,newcall("_gerepilemany",gpm));
+  cseq=newnode(Fassign,var,newcall(next,vprime));
+  tree[cseq].m=Mparens;
+  aseq=addseqright(aseq,newcall("_whilenull",newnode(Flistarg,cseq,seq)));
+  if (lab.go)
+    aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
+  makeblocks3(bl0,bl1,bl2,n,bseq,aseq,ret,savx,savx);
+  copyctx(sav0,block+bl0);
+  restorectx(sav0);
+}
+
 /*
    The 6 operations which order is very important are
 are:
@@ -189,7 +360,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
   Clabel_t lab;
   switch(gp->spec)
   {
-    case GPaddhelp:
+  case GPaddhelp:
     {  /*prototype vSs*/
       gpfunc *gp;
       int d, nf;
@@ -200,9 +371,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       tree[d].f=Fconst;
       nf=getfunc(entryname(d)); gp=lfunc+nf;
       d=detag(arg[1]);
-      if (tree[d].f!=Fentry && !is_const(d,CSTstr))
-        die(n,"Incorrect help text in addhelp");
-      gp->proto.help=entryname(d);
+      gp->proto.help=d;
     }
     break;
   case GPgrandO:
@@ -212,7 +381,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     genblock(y,n);
     break;
   case GPif:
-    nb=genlistargs(n,arg,2,3);
+    nb=genlistargs(n,arg,2,STACKSZ);
     if (nb==3 && arg[2]==GNOARG)
       /*Some people type 'if(x!=0,print(x),)'*/
       nb--;
@@ -235,11 +404,19 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     else
     {
       if (p>=0 && tree[p].f!=Fseq)
+        newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
+      if (nb>=4)
       {
-        if(nb==3)
-          newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
-        else
-          newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
+        arg[2] = newcall("if", stacktoargs(arg+2,nb-2));
+        nb = 3;
+      }
+      if (arg[0]>=0 && tree[arg[0]].f==Fseq)
+      {
+        int pred;
+        newdecl((1<<Cauto)|(1<<Cimmutable),Gbool,GNIL,&pred);
+        aseq = geninsertvar(arg[0], pred);
+        genblock(aseq,n);
+        arg[0] = newleaf(pred);
       }
       genblock(arg[0],n);
       arg[1]=geninsertvar(arg[1],ret);
@@ -249,7 +426,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         arg[2]=geninsertvar(arg[2],ret==-1?-1:newleaf(ret));
         makesubblock(arg[2]);
       }
-      aseq=newnode(tree[n].f,tree[n].x,listtoseq(arg,nb));
+      aseq=addseqright(aseq,newnode(tree[n].f,tree[n].x,stacktoargs(arg,nb)));
     }
     makeblock(newblock(),n,aseq,ret,s_ctx.n);
     break;
@@ -297,6 +474,9 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       case 'l':
         functype(*gp)=Gsmall;
         break;
+      case 'u':
+        functype(*gp)=Gusmall;
+        break;
       case 'i':
         functype(*gp)=Gsmall_int;
         break;
@@ -328,18 +508,18 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     bl2=newblock();
     if (autogc)
       gpm=gengerepile(bl2);
-    genequaltosmall(binf,bsup,&tv);
+    genequaltosmall(binf,bsup,&tv,Ggen);
     pushvar(var,1<<Cuser,tv,-1);
-    lab=makesubblock_label(seq);
+    lab=makesubblock_label(seq,Lbreak);
     if (lab.ne)
       seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     if (autogc)
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-    arg[0]=newnode(Faffect,var,vinf);
+    arg[0]=newnode(Fassign,var,vinf);
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=seq;
     arg[3]=newopcall(OPpp,newleaf(var),-1);
-    aseq=newnode(tree[n].f,tree[n].x,listtoseq(arg,4));
+    aseq=newnode(tree[n].f,tree[n].x,stacktoargs(arg,4));
     if (lab.go)
       aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
     makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
@@ -362,18 +542,18 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       pushvar(vfvnext,(1<<Cuser)|(1<<Cconst),Gfunc_GG,-1);
       pushvar(var,(1<<Cuser)|(1<<Ccompo),Gvec,newcall("_forvec_start",
             newnode(Flistarg,newnode(Flistarg,newnode(Flistarg,
-              newleaf(vinf),newleaf(flag)),
-              newnode(Frefarg,newnode(Fentry,newentry("fv_data"),-1),-1)),
+                  newleaf(vinf),newleaf(flag)),
+                newnode(Frefarg,newnode(Fentry,newentry("fv_data"),-1),-1)),
               newnode(Frefarg,newnode(Fentry,newentry("fv_next"),-1),-1))));
       savx=s_ctx.n;
       bl1=newblock();
       bl2=newblock();
       if (autogc)
         gpm=gengerepile(bl2);
-      next = newnode(Faffect,newleaf(var),newnode(Ftag,
-          newcall("_call_GG",newnode(Flistarg,newnode(Flistarg,
-              newleaf(vfvnext),newleaf(vfvdat)),newleaf(var))),Gvec));
-      lab=makesubblock_label(seq);
+      next = newnode(Fassign,newleaf(var),newnode(Ftag,
+            newcall("_call_GG",newnode(Flistarg,newnode(Flistarg,
+                  newleaf(vfvnext),newleaf(vfvdat)),newleaf(var))),Gvec));
+      lab=makesubblock_label(seq,Lbreak);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
@@ -382,7 +562,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       arg[1]=newleaf(var);
       arg[3]=next;
       arg[2]=seq;
-      aseq=newcall("for",listtoseq(arg,4));
+      aseq=newcall("for",stacktoargs(arg,4));
       if (lab.go)
         aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
       makeblocks3(bl0,bl1,bl2,n,bseq,aseq,ret,sav0,savx);
@@ -407,20 +587,20 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       newdecl(0,Gsmall,-1,&vloop);
       if (tv==Gnotype) tv=Gint;
       pushvar(var,1<<Cuser,tv,-1);
-      lab=makesubblock_label(seq);
+      lab=makesubblock_label(seq,Lbreak);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
       seq=addseqleft(
-          newnode(Faffect,newleaf(var),
+          newnode(Fassign,newleaf(var),
             newnode(Ftag,
               newcoeff(newleaf(vdiv),newleaf(vloop),-1),Gint)),seq);
-      arg[0]=newnode(Faffect,vloop,newsmall(1));
+      arg[0]=newnode(Fassign,vloop,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(vloop),newcall("length",vdiv));
       arg[2]=seq;
       arg[3]=newopcall(OPpp,newleaf(vloop),-1);
-      aseq=newnode(tree[n].f,newentry("for"),listtoseq(arg,4));
+      aseq=newnode(tree[n].f,newentry("for"),stacktoargs(arg,4));
       if (lab.go)
         aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
       makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
@@ -428,7 +608,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     }
   case GPforstep:
     {
-      int bstep;
+      int bstep,lstep,sstep;
       int vstep,vindex,vsign;
       int dosign,isvec;
       nb=genlistargs(n,arg,4,4);
@@ -442,14 +622,29 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       bseq=addseqright(bseq,geninsertvar(bstep,newleaf(vstep)));
       if (guesstype(bstep)==Gvec)
       {
-        isvec=1;
-        newdecl(0,Gsmall,-1,&vindex);
-        dosign=0;
+        int sarg[3];
+        isvec=1; dosign=1;
+        newdecl((1<<Cauto)|(1<<Cconst),Gsmall,-1,&vindex);
+        newdecl((1<<Cauto)|(1<<Cconst),Gsmall,-1,&lstep);
+        bseq=addseqright(bseq,geninsertvar(newcall("length",newleaf(vstep)),newleaf(lstep)));
+        sarg[0] = newnode(Fassign,newleaf(vindex),newsmall(1));
+        sarg[1] = newleaf(lstep);
+        sarg[2] = newnode(Fmatcoeff,newleaf(vstep),
+            newnode(Fmatrix,newnode(Frange,newleaf(vindex),GNOARG),-1));
+        sstep = newcall("sum", stacktoargs(sarg,3));
       }
       else
       {
         isvec=0;
         dosign=tree[bstep].f!=Fsmall;
+        if (dosign)
+          sstep = newleaf(vstep);
+      }
+      if (dosign)
+      {
+        newdecl((1<<Cauto)|(1<<Cconst),Gbool,-1,&vsign);
+        sstep = newopcall(OPg,sstep, newsmall(0));
+        bseq=addseqright(bseq,geninsertvar(sstep,newleaf(vsign)));
       }
       genblock(bseq,-1);
       savx=s_ctx.n;
@@ -460,65 +655,69 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (tv==Gnotype)
         tv=Ggen;
       pushvar(var,1<<Cuser,tv,-1);
-      if (dosign)
-        newdecl(0,Gbool,newopcall(OPg,newleaf(vstep),0),&vsign);
-      lab=makesubblock_label(seq);
+      lab=makesubblock_label(seq,Lbreak);
       if (lab.ne)
         seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
-      arg[0]=newnode(Faffect,var,vinf);
-      if (isvec)
+      arg[0]=newnode(Fassign,var,vinf);
+      if ( !dosign )
       {
-        arg[4]=vindex;
-        arg[5]=newcall("length",vstep);
-        arg[3]=newopcall(OPle,newleaf(var),vsup);
-        arg[2]=newopcall(OPpe,var,newcoeff(vstep,vindex,-1));
+        if ( tree[bstep].x>0 )
+          arg[3]=newopcall(OPle,newleaf(var),vsup);
+        else
+          arg[3]=newopcall(OPge,newleaf(var),vsup);
+        nb=4;
+      }
+      else
+      {
+        arg[3]=newleaf(vsign);
+        arg[4]=newopcall(OPle,newleaf(var),vsup);
+        arg[5]=newopcall(OPge,newleaf(var),vsup);
         nb=6;
       }
-      else
+      if (isvec)
       {
-        if ( !dosign )
-        {
-          if ( tree[bstep].x>0 )
-            arg[3]=newopcall(OPle,newleaf(var),vsup);
-          else
-            arg[3]=newopcall(OPge,newleaf(var),vsup);
-          nb=4;
-        }
-        else
-        {
-          arg[3]=newleaf(vsign);
-          arg[4]=newopcall(OPle,newleaf(var),vsup);
-          arg[5]=newopcall(OPge,newleaf(var),vsup);
-          nb=6;
-        }
-        arg[2]=newopcall(OPpe,var,vstep);
+        arg[2]=newopcall(OPpe,var,newcoeff(vstep,vindex,-1));
+        arg[6]=vindex;
+        arg[7]=newleaf(lstep);
+        nb=8;
       }
+      else
+        arg[2]=newopcall(OPpe,var,vstep);
       arg[1]=seq;
       if (isvec)
-        aseq=newcall("forstepvec",listtoseq(arg,nb));
+        aseq=newcall("_forstepvec",stacktoargs(arg,nb));
       else
-        aseq=newnode(tree[n].f,tree[n].x,listtoseq(arg,nb));
+        aseq=newnode(tree[n].f,tree[n].x,stacktoargs(arg,nb));
       if (lab.go)
         aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
       makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
     }
     break;
-  case GPforprime:
   case GPprodeuler:
+    nb=genlistargs(n,arg,3,3);
+    seq = arg[2];
+    newdecl(0,Ggen,-1,&ret);
+    savx=s_ctx.n;
+    bl1=newblock();
+    arg[2]=geninsertvarop(seq,ret,OPme);
+    aseq=newnode(Fseq,
+            geninsertvar(newnode(Fconst,newsmallrealvalue(1),-1),newleaf(ret)),
+                 newcall("forprime",stacktoargs(arg,3)));
+    genblock(aseq,-1);
+    makeblock(bl1,n,aseq,ret,savx);
+    break;
+  case GPforprime:
     nb=genlistargs(n,arg,3,3);
     genequal(arg[0],gp->gpname,&var,&binf,&tv);
     bsup=arg[1];seq=arg[2];
     newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vinf);
     bseq=addseqright(bseq,geninsertvar(binf,newleaf(vinf)));
-    newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vsup);
-    bseq=addseqright(bseq,geninsertvar(bsup,newleaf(vsup)));
-    if (gp->spec==GPprodeuler)
+    if (bsup!=GNOARG || FC_forprime_init<0)
     {
-      newdecl(0,Ggen,-1,&ret);
-      aseq=addseqright(aseq,geninsertvar(
-            newnode(Fconst,newsmallrealvalue(1),-1),newleaf(ret)));
+      newdecl((1<<Cauto)|(1<<Cconst),Gempty,-1,&vsup);
+      bseq=addseqright(bseq,geninsertvar(bsup,newleaf(vsup)));
     }
     genblock(bseq,-1);
     savx=s_ctx.n;
@@ -526,38 +725,40 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     bl2=newblock();
     if (autogc)
       gpm=gengerepile(bl2);
-    if (tv==Gnotype)
-      tv=Gsmall;
-    pushvar(var,1<<Cuser,tv,FC_forprime_next>=0?newsmall(0):-1);
+    genequaltosmall(binf,bsup,&tv,FC_forprime_init>=0?Gint:Gsmall);
+    pushvar(var,1<<Cuser,tv,FC_forprime_init>=0?-1:newsmall(0));
     vprime=newnode(Fentry,newentry("primepointer"),-1);
-    pushvar(vprime,1<<Cuser,Gbptr,newcall("_diffptr",-1));
-    if (gp->spec==GPprodeuler)
+    pushvar(vprime,1<<Cuser,Gforprime,FC_forprime_init>=0?-1:newcall("_diffptr",-1));
+    lab=makesubblock_label(seq,Lbreak);
+    if (lab.ne)
+      seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
+    if (autogc)
+      seq=addseqright(seq,newcall("_gerepilemany",gpm));
+    if (FC_forprime_init>=0)
     {
-      seq=geninsertvarop(seq,ret,OPme);
-      makesubblock_nolabel(seq);
-      lab.go=0;
+      int args[3],bseq;
+      int nbs=3;
+      args[0]=vprime; args[1]=vinf;
+      if (bsup==GNOARG) nbs=2; else args[2]=vsup;
+      aseq=addseqleft(newcall((tv==Gsmall?"_u_forprime_init":"_forprime_init"),
+            stacktoargs(args,nbs)),aseq);
+      bseq=newnode(Fassign,var,newcall(tv==Gsmall?"_u_forprime_next":"_forprime_next_",vprime));
+      tree[bseq].m=Mparens;
+      aseq=addseqright(aseq,newcall("_whilenull",newnode(Flistarg,bseq,seq)));
     }
     else
     {
-      lab=makesubblock_label(seq);
-      if (lab.ne)
-        seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
-    }
-    if (autogc)
-      seq=addseqright(seq,newcall("_gerepilemany",gpm));
-    /*if (vsup>maxprime()) err(primer1);*/
-    aseq=addseqleft(newcall("if",newnode(Flistarg,
+      /*if (vsup>maxprime()) err(primer1);*/
+      aseq=addseqleft(newcall("if",newnode(Flistarg,
               newopcall(OPg,vsup,newcall("_maxprime",-1)),
               newcall("_err_primes",-1)))
-              ,aseq);
-    if (tree[binf].f!=Fsmall || tree[binf].x>2)
-      /*if(var<vinf) continue;*/
-      seq=addseqleft(newcall("if",newnode(Flistarg,
-              newopcall(OPl,newleaf(var),vinf),
-              newcall("next",-1)
-              )),seq);
-    if (FC_forprime_next>=0)
-    {
+          ,aseq);
+      if (tree[binf].f!=Fsmall || tree[binf].x>2)
+        /*if(var<vinf) continue;*/
+        seq=addseqleft(newcall("if",newnode(Flistarg,
+                newopcall(OPl,newleaf(var),vinf),
+                newcall("next",-1)
+                )),seq);
       seq=addseqleft(newcall("if",newnode(Flistarg,
               newopcall(OPg,newleaf(var),vsup),
               newcall("break",-1)
@@ -567,18 +768,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
               newleaf(var),newleaf(vprime))),seq);
       aseq=addseqright(aseq,newcall("_doloop",seq));
     }
-    else
-    {
-      arg[0]=newnode(Faffect,var,newsmall(2));
-      /*var<=vsup*/
-      arg[1]=newopcall(OPle,newleaf(var),newleaf(vsup));
-      arg[2]=seq;
-      /*var+=*++primepointer*/
-      arg[3]=geninsertvarop(newopcall(OPpp,newleaf(vprime),-1),
-          newleaf(var),OPpe);
-      aseq=addseqright(aseq,newcall("for",listtoseq(arg,4)));
-    }
-    if (gp->spec==GPforprime && lab.go)
+    if (lab.go)
       aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
     makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
     break;
@@ -594,7 +784,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     newdecl(0,Ggen,-1,&ret);
     if (nb==3)
       aseq=addseqright(aseq,geninsertvar(newsmall(gp->spec==GPsum?0:1),
-                                         newleaf(ret)));
+            newleaf(ret)));
     else
       bseq=addseqright(bseq,geninsertvar(arg[3],newleaf(ret)));
     genblock(bseq,-1);
@@ -603,17 +793,17 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     bl2=newblock();
     if (autogc)
       gpm=gengerepile(bl2);
-    genequaltosmall(binf,bsup,&tv);
+    genequaltosmall(binf,bsup,&tv,Gint);
     pushvar(var,1<<Cuser,tv,-1);
     seq=geninsertvarop(seq,ret,(gp->spec==GPsum?OPpe:OPme));
-    makesubblock_nolabel(seq);
+    makesubblock_label(seq,Lnobrk);
     if (autogc)
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
-    arg[0]=newnode(Faffect,var,vinf);
+    arg[0]=newnode(Fassign,var,vinf);
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=seq;
     arg[3]=newopcall(OPpp,newleaf(var),-1);
-    aseq=addseqright(aseq,newcall("for",listtoseq(arg,4)));
+    aseq=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
     makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
     break;
   case GPsumdiv:
@@ -638,18 +828,18 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       if (tv==Gnotype) tv=Gint;
       pushvar(var,1<<Cuser,tv,-1);
       seq=geninsertvarop(seq,ret,OPpe);
-      makesubblock_nolabel(seq);
+      makesubblock_label(seq,Lnobrk);
       if (autogc)
         seq=addseqright(seq,newcall("_gerepilemany",gpm));
       seq=addseqleft(
-          newnode(Faffect,newleaf(var),
+          newnode(Fassign,newleaf(var),
             newnode(Ftag,
               newcoeff(newleaf(vdiv),newleaf(vloop),-1),Gint)),seq);
-      arg[0]=newnode(Faffect,vloop,newsmall(1));
+      arg[0]=newnode(Fassign,vloop,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(vloop),newcall("length",vdiv));
       arg[2]=seq;
       arg[3]=newopcall(OPpp,newleaf(vloop),-1);
-      aseq=newnode(tree[n].f,newentry("for"),listtoseq(arg,4));
+      aseq=newnode(tree[n].f,newentry("for"),stacktoargs(arg,4));
       makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
       break;
     }
@@ -661,14 +851,14 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     seq=arg[1];
     if(autogc)
       gpm=gengerepile(bl1);
-    genblock_nolabel(arg[0],n);
-    lab=makesubblock_label(seq);
+    genblock_label(arg[0],n,Lnobrk);
+    lab=makesubblock_label(seq,Lbreak);
     if (lab.ne)
       seq=newseq(seq,newcall("_label",newsmall(lab.ne)));
     if (autogc)
       seq=addseqright(seq,newcall("_gerepilemany",gpm));
     arg[1]=seq;
-    aseq=newnode(tree[n].f,tree[n].x,listtoseq(arg,nb));
+    aseq=newnode(tree[n].f,tree[n].x,stacktoargs(arg,nb));
     if (lab.go)
       aseq=newseq(aseq,newcall("_label",newsmall(lab.go)));
     makeblock(bl1,n,aseq,ret,savx);
@@ -693,14 +883,14 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     else
       newdecl(0,Gsmall,-1,&var);
     aseq=addseqright(aseq,newgetg(ret,newleaf(vsup),
-        gp->spec==GPvector?"t_VEC":gp->spec==GPvectorv?"t_COL":"t_VECSMALL"));
-    arg[0]=newnode(Faffect,var,newsmall(1));
+          gp->spec==GPvector?"t_VEC":gp->spec==GPvectorv?"t_COL":"t_VECSMALL"));
+    arg[0]=newnode(Fassign,var,newsmall(1));
     arg[1]=newopcall(OPle,newleaf(var),vsup);
     arg[2]=newcoeff(newleaf(ret),newleaf(var),-1);
     arg[2]=geninsertvar(seq,arg[2]);
-    makesubblock_nolabel(arg[2]);
+    makesubblock_label(arg[2],Lnobrk);
     arg[3]=newopcall(OPpp,newleaf(var),-1);
-    aseq=addseqright(aseq,newcall("for",listtoseq(arg,4)));
+    aseq=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
     makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
     break;
   case GPmatrix:
@@ -737,19 +927,19 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       else
         seq=arg[4];
       aseq=newgetg(newcoeff(newleaf(ret),newleaf(var1),-1),
-                     newleaf(vsup2),"t_COL");
-      arg[0]=newnode(Faffect,var2,newsmall(1));
+          newleaf(vsup2),"t_COL");
+      arg[0]=newnode(Fassign,var2,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(var2),vsup2);
       arg[3]=newopcall(OPpp,newleaf(var2),-1);
       arg[2]=newcoeff(newleaf(ret),newleaf(var2),newleaf(var1));
       arg[2]=geninsertvar(seq,arg[2]);
-      makesubblock_nolabel(arg[2]);
-      arg[2]=addseqright(aseq,newcall("for",listtoseq(arg,4)));
+      makesubblock_label(arg[2],Lnobrk);
+      arg[2]=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
       aseq=newgetg(ret,newleaf(vsup1),"t_MAT");
-      arg[0]=newnode(Faffect,var1,newsmall(1));
+      arg[0]=newnode(Fassign,var1,newsmall(1));
       arg[1]=newopcall(OPle,newleaf(var1),vsup1);
       arg[3]=newopcall(OPpp,newleaf(var1),-1);
-      aseq=addseqright(aseq,newcall("for",listtoseq(arg,4)));
+      aseq=addseqright(aseq,newcall("for",stacktoargs(arg,4)));
       makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
     }
     break;
@@ -760,7 +950,7 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
       int flag=1<<Cuser;
       int mint=(tree[p].f==Ftag)?tree[p].y:-1;
       if (gp->spec==GPglobal)
-        flag|=1<<Cglobal;
+        flag|=(1<<Cglobal)|(1<<Cfunction);
       savx=s_ctx.n;
       (void)genblockdeclaration(y,n,flag,mint,&aseq);
       makeblock(newblock(),n,aseq,-1,s_ctx.n);
@@ -768,49 +958,57 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
     }
     break;
   case GPreturn:
-    if (y==GNOARG || y<0)
-      y=tree[n].y=GNIL;
-    else
-      genblock(y,n);
-    /* If we don't do garbage collecting, don't bother adding blocks.*/
-    if (!autogc) break;
-    if (tree[y].f==Fentry)
     {
-      var=y;
-      bseq=-1;
-      y=newleaf(var);
+      int iferr_reset=0;
+      if (y==GNOARG || y<0)
+        y=tree[n].y=GNIL;
+      else
+        genblock(y,n);
+      iferr_reset=check_labels(s_label.n,Liferr)==1;
+      /* If we don't do garbage collecting, don't bother adding blocks.*/
+      if (!autogc && !iferr_reset) break;
+      if (tree[y].f==Fentry && !(ctxstack[getvarerr(y)].flag&(1<<Cglobal)))
+      {
+        var=y;
+        bseq=-1;
+        y=newleaf(var);
+      }
+      else if (tree[y].f==Fblock)
+      {
+        /*Use the block value variable as the return variable*/
+        var=block[tree[y].x].ret;
+        bseq=y;
+        y=newleaf(var);
+      }
+      else if (tree[y].f!=Fsmall)
+      {
+        /*Use the returned variable as the return variable*/
+        newdecl((1<<Cauto),Gempty,-1,&var);
+        bseq=geninsertvar(y,var);
+        y=newleaf(var);
+      }
+      bl1=newblock();
+      bl2=newblock();
+      savx=s_ctx.n;
+      block[bl2].gc|=(1<<GCneeded)|(1<<GCglobal)|(1<<GCreturn);
+      block[bl2].egc=var;
+      if (autogc)
+      {
+        arg[0]=newnode(Fentry,newentry("ltop"),-1);
+        arg[1]=GNOARG;
+        arg[2]=newsmall(bl2);
+        gpm=stacktoargs(arg,3);
+        aseq=addseqright(aseq,newcall("_gerepilemany",gpm));
+      }
+      if (iferr_reset)
+        aseq=addseqright(aseq,newcall("_iferr_CATCH_reset",newsmall(0)));
+      aseq=addseqright(aseq,newnode(tree[n].f,tree[n].x,y));
+      makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
+      break;
     }
-    else if (tree[y].f==Fblock)
-    {
-      /*Use the block value variable as the return variable*/
-      var=block[tree[y].x].ret;
-      bseq=y;
-      y=newleaf(var);
-    }
-    else if (tree[y].f!=Fsmall)
-    {
-      /*Use the returned variable as the return variable*/
-      newdecl((1<<Cauto),Gempty,-1,&var);
-      bseq=geninsertvar(y,var);
-      y=newleaf(var);
-    }
-    bl1=newblock();
-    bl2=newblock();
-    savx=s_ctx.n;
-    block[bl2].gc|=(1<<GCneeded)|(1<<GCglobal)|(1<<GCreturn);
-    block[bl2].egc=var;
-    arg[0]=newnode(Fentry,newentry("ltop"),-1);
-    arg[1]=GNOARG;
-    arg[2]=newsmall(bl2);
-    gpm=listtoseq(arg,3);
-    aseq=addseqright(aseq,newcall("_gerepilemany",gpm));
-    aseq=addseqright(aseq,newnode(tree[n].f,tree[n].x,y));
-    makeblocks(bl1,bl2,n,bseq,aseq,ret,savx);
-    break;
   case GPbreak:
   case GPnext:
     {
-      int k,l;
       static int labbrk=0;
       if (y>=0 && y!=GNOARG)
       {
@@ -823,13 +1021,11 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
         nb=1;
         tree[n].y=GNOARG;
       }
-      l=s_label.n;
-      k=nb;
-      while(k--)
-        if (!l || !label[--l].ok)
-          die(n,"%s not allowed here",gp->gpname);
+      if (check_labels(nb,Lnobrk))
+        die(n,"%s not allowed here",gp->gpname);
       if (nb>1)
       {
+        int l = s_label.n-nb;
         if (gp->spec==GPbreak)
         {
           if (label[l].go==0)
@@ -843,8 +1039,49 @@ void genblockfuncspec(int n, int p, gpfunc *gp)
           tree[y].x=labbrk;
         }
       }
+      if (check_labels(nb,Liferr))
+      {
+        int ie = newsmall(get_iferr(nb));
+        aseq=addseqleft(newcall("_iferr_CATCH_reset",ie),newleaf(n));
+        makeblock(newblock(),n,aseq,-1,s_ctx.n);
+      }
     }
     break;
+  case GPiferr:
+    {
+      Clabel_t lab;
+      nb=genlistargs(n,arg,3,4);
+      checkisvar(arg[1],gp->gpname,&var,&tv);
+      seq=arg[0];
+      bseq=arg[2];
+      if (tv==Gnotype)
+      {
+        tv=Gerror;
+        if (tv<0) die(n,"iferr not supported by your GP version");
+      }
+      if (p>=0 && tree[p].f!=Fseq)
+        newdecl((1<<Cauto)|(1<<Cimmutable),Gvoid,GNIL,&ret);
+      seq=geninsertvar(seq,ret);
+      bseq=geninsertvar(bseq,ret);
+      lab=makesubblock_label(seq,Liferr);
+      savx=s_ctx.n;
+      pushvar(var,1<<Cuser,tv,newcall("_iferr_error",-1));
+      if(nb==4)
+      {
+        int pred = newcall("if", newnode(Flistarg,
+                      newnode(Flistarg,arg[3],GNOARG),
+                      newcall("_iferr_rethrow",var)));
+        bseq = addseqleft(pred,bseq);
+      }
+      makesubblock_ctx(bseq,savx);
+      arg[0]=newcall("_iferr_CATCH", newsmall(lab.ie));
+      arg[1]=bseq;
+      arg[2]=seq;
+      arg[3]=newcall("_iferr_ENDCATCH",newsmall(lab.ie));
+      aseq=addseqright(aseq,newcall("iferr",stacktoargs(arg,4)));
+      makeblock(newblock(),n,aseq,ret,s_ctx.n);
+      break;
+    }
   default:
     /*treat as a normal function*/
     genblock(tree[n].y,n);
@@ -890,6 +1127,19 @@ int gentypefuncspec(int n, gpfunc *gp)
     if ((tree[arg[1]].m&(1<<Mterm)))
       tree[n].m|=(1<<Msidef);
     return tree[arg[1]].t;
+  case GPwhilenull:
+    nb=genlistargs(n,arg,2,2);
+    gentypemode(nb,arg,&tree[n].m);
+    tree[n].m&=~(1<<Mterm);
+    return Gvoid;
+  case GPiferr:
+    nb=genlistargs(n,arg,4,4);
+    gentypemode(4,arg,&tree[n].m);
+    tree[arg[1]].m|=1<<Muntil;
+    tree[arg[2]].m|=1<<Muntil;
+    tree[arg[3]].m|=1<<Muntil;
+    tree[n].m&=~(1<<Mterm);
+    return typemax[tree[arg[1]].t][tree[arg[2]].t];
   case GPdoloop:
     gentype(y);
     tree[n].m=tree[y].m&MODHERIT;
@@ -901,10 +1151,10 @@ int gentypefuncspec(int n, gpfunc *gp)
     return Gvoid;
   case GPforstep:
   case GPforstepvec:
-    nb=genlistargs(n,arg,4,7);
+    nb=genlistargs(n,arg,4,8);
     gentypemode(nb,arg,&tree[n].m);
     tree[n].m&=~(1<<Mterm);
-   return Gvoid;
+    return Gvoid;
   case GPuntil:
   case GPwhile:
     nb=genlistargs(n,arg,2,2);
@@ -917,8 +1167,6 @@ int gentypefuncspec(int n, gpfunc *gp)
       int a = arg[1];
       tree[n].m=((tree[arg[0]].m|tree[arg[1]].m)&MODHERIT);
       tree[a].m|=(1<<Muntil);
-      if (tree[a].f==Fblock)
-        tree[tree[a].y].m|=(1<<Muntil);
     }
     return Gvoid;
   case GPreturn:
@@ -985,7 +1233,7 @@ void gencopyarg(FILE *fout)
       ctxvar *v=fc->c+i;
       long j;
       for(j=0;j<fc->v.n;j++)
-        if (fc->var[j].f==AFaffectcompo && fc->var[j].idx==i+fc->savb-fc->s.n )
+        if (fc->var[j].f==AFassigncompo && fc->var[j].idx==i+fc->savb-fc->s.n )
         {
           genindent(fout);
           fprintf(fout,"%s = ",v->cvar);
@@ -1034,24 +1282,7 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       break;
     case GPif:
       nb=genlistargs(n,arg,2,3);
-#if 0
-      /*generate nice `? :' */
-      if (t!=Gvoid && tree[arg[1]].f!=Fseq && (nb==2 || tree[arg[2]].f!=Fseq) )
-      {
-        fprintf(fout,"((");
-        genbool(fout,arg[0],tree[arg[0]].t);
-        fprintf(fout,")?");
-        gencast(fout,arg[3],t);
-        fprintf(fout,":");
-        if (nb==5)
-          gencast(fout,arg[4],t);
-        else/*the type of left and right must be equal.*/
-          gensmallval(fout,0,t);
-        fprintf(fout,")");
-      }
-#endif
-      genindent(fout);
-      fprintf(fout,"if (");
+      genindentline(fout,"if (");
       gencast(fout,arg[0],Gbool);
       fprintf(fout,")\n");
       if (nb==2)
@@ -1059,17 +1290,34 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       else
       {
         genbrace(fout,arg[1]);
-        genindent(fout);
-        fprintf(fout,"else\n");
+        genindentline(fout,"else\n");
         genbrace(fout,arg[2]);
       }
       break;
+    case GPiferr:
+      nb=genlistargs(n,arg,4,4);
+      genindent(fout);
+      gencode(fout,arg[0]);
+      fprintf(fout,"\n");
+      genbrace(fout,arg[1]);
+      fprintf(fout," pari_TRY\n");
+      genbrace(fout,arg[2]);
+      fprintf(fout," ");
+      gencode(fout,arg[3]);
+      fprintf(fout,"\n");
+      break;
+    case GPwhilenull:
+      nb=genlistargs(n,arg,2,2);
+      genindentline(fout,"while (");
+      genparens(fout,arg[0]);
+      fprintf(fout,")\n");
+      genbrace(fout,arg[1]);
+      break;
     case GPwhile:
       nb=genlistargs(n,arg,2,2);
-      genindent(fout);
       if (tree[arg[0]].f!=Fseq && tree[arg[0]].f!=Fblock)
       {
-        fprintf(fout,"while (");
+        genindentline(fout,"while (");
         gencast(fout,arg[0],Gbool);
         fprintf(fout,")\n");
         genbrace(fout,arg[1]);
@@ -1083,40 +1331,35 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
           y=block[x].ret;
           x=tree[arg[0]].y;
         }
-        fprintf(fout,"for(;;)\n");
-        genindent(fout);
-        fprintf(fout,"{\n");
+        genindentline(fout,"for(;;)\n");
+        genindentline(fout,"{\n");
         indent++;
         genindentseq(fout,x);
-        gencodeg(fout,x);
+        gencode(fout,x);
         gensemicolon(fout,x);
         if (tree[y].f==Fblock)
         {
-          gencodeg(fout,y);
+          gencode(fout,y);
           y=block[tree[y].x].ret;
         }
-        genindent(fout);
-        fprintf(fout,"if (");
+        genindentline(fout,"if (");
         gencast(fout,y,Gnegbool);
         fprintf(fout,")\n");
         indent++;
-        genindent(fout);
-        fprintf(fout,"break;\n");
+        genindentline(fout,"break;\n");
         indent--;
         genindentseq(fout,arg[1]);
-        gencodeg(fout,arg[1]);
+        gencode(fout,arg[1]);
         gensemicolon(fout,arg[1]);
         indent--;
-        genindent(fout);
-        fprintf(fout,"}\n");
+        genendbrace(fout,n);
       }
       break;
     case GPuntil:
       nb=genlistargs(n,arg,2,2);
-      genindent(fout);
       if ( tree[arg[0]].f!=Fseq && tree[arg[0]].f!=Fblock)
       {
-        fprintf(fout,"do\n");
+        genindentline(fout,"do\n");
         genbrace(fout,arg[1]);
         fprintf(fout," while(");/*KB style*/
         gencast(fout,arg[0],Gnegbool);
@@ -1131,38 +1374,34 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
           y=block[x].ret;
           x=tree[arg[0]].y;
         }
-        fprintf(fout,"for(;;)\n");
-        genindent(fout);
-        fprintf(fout,"{\n");
+        genindentline(fout,"for(;;)\n");
+        genindentline(fout,"{\n");
         indent++;
         genindentseq(fout,arg[1]);
-        gencodeg(fout,arg[1]);
+        gencode(fout,arg[1]);
         gensemicolon(fout,arg[1]);
         genindentseq(fout,x);
-        gencodeg(fout,x);
+        gencode(fout,x);
         gensemicolon(fout,x);
-        genindent(fout);
         if (tree[y].f==Fblock)
         {
-          gencodeg(fout,y);
+          genindent(fout);
+          gencode(fout,y);
           y=block[tree[y].x].ret;
         }
-        fprintf(fout,"if (");
+        genindentline(fout,"if (");
         gencast(fout,y,Gbool);
         fprintf(fout,")\n");
         indent++;
-        genindent(fout);
-        fprintf(fout,"break;\n");
+        genindentline(fout,"break;\n");
         indent--;
         indent--;
-        genindent(fout);
-        fprintf(fout,"}\n");
+        genendbrace(fout,n);
       }
       break;
     case GPfor:
       nb=genlistargs(n,arg,4,4);
-      genindent(fout);
-      fprintf(fout,"for (");
+      genindentline(fout,"for (");
       if (arg[0]!=GNOARG)
         gencode(fout,arg[0]);
       else
@@ -1177,14 +1416,12 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
       genbrace(fout,arg[2]);
       break;
     case GPdoloop:
-      genindent(fout);
-      fprintf(fout,"for (;;)\n");
+      genindentline(fout,"for (;;)\n");
       genbrace(fout,y);
       break;
     case GPforstep:
       nb=genlistargs(n,arg,4,6);
-      genindent(fout);
-      fprintf(fout,"for (");
+      genindentline(fout,"for (");
       gencode(fout,arg[0]);
       fprintf(fout,"; ");
       if ( nb==4 )
@@ -1213,22 +1450,25 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     }
 #endif
   case GPforstepvec:
-    nb=genlistargs(n,arg,6,6);
-    genindent(fout);
-    fprintf(fout,"for (");
-    gencode(fout,arg[4]);
+    nb=genlistargs(n,arg,8,8);
+    genindentline(fout,"for (");
+    gencode(fout,arg[6]);
     fprintf(fout," = 0, ");
     gencode(fout,arg[0]);
     fprintf(fout,"; ");
     gencode(fout,arg[3]);
+    fprintf(fout,"?");
+    gencode(fout,arg[4]);
+    fprintf(fout,":");
+    gencode(fout,arg[5]);
     fprintf(fout,"; ");
-    gencode(fout,arg[4]);
+    gencode(fout,arg[6]);
     fprintf(fout," = ");
-    gencode(fout,arg[4]);
+    gencode(fout,arg[6]);
     fprintf(fout," == ");
-    gencast(fout,arg[5],Gsmall);
+    gencast(fout,arg[7],Gsmall);
     fprintf(fout," ? 1 :");
-    gencode(fout,arg[4]);
+    gencode(fout,arg[6]);
     fprintf(fout," + 1, ");
     gencode(fout,arg[2]);
     fprintf(fout,")\n");
@@ -1242,7 +1482,7 @@ void genentryspec(FILE *fout, int n, gpfunc *gp)
     for(i=1;i<nb;i++)
     {
       genindentseq(fout,arg[i]);
-      gencodeg(fout,arg[i]);
+      gencode(fout,arg[i]);
       gensemicolon(fout,arg[i]);
     }
     break;
