@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2000-2005  The PARI group.
+Copyright (C) 2000-2013  The PARI group.
 
 This file is part of the GP2C package.
 
@@ -10,7 +10,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
 
 #include "config.h"
 #include <stdlib.h>
@@ -31,11 +31,12 @@ static int newfunc(const char *gpname)
   f->proto.cname=gpname;
   f->proto.code=NULL;
   f->proto.origin=NULL;
-  f->proto.help=NULL;
+  f->proto.help=-1;
   f->node=newnode(Ffun,nf,-1);
   functype(*f)=Gempty;
   funcmode(*f)=0;
   f->spec=GPpari;
+  f->iter=NULL;
   f->wrap=NULL;
   f->dsc=NULL;
   f->user=NULL;
@@ -108,6 +109,15 @@ int findfunctype(const char *s)
   return da->type;
 }
 
+int findfunctypeopt(const char *s)
+{
+  gpdescarg *da;
+  int nf=findfuncdescopt(s);
+  if (nf<0) return -1;
+  da=descfindrules(0,NULL,lfunc+nf);
+  return da->type;
+}
+
 int getfunc(const char *gpname)
 {
   int r=findfunction(gpname);
@@ -140,6 +150,9 @@ int genarg(int nerr, FILE *fout, char c, int n, PPproto pr)
   case 'G':
     gencast(fout,n,Ggen);
     return 1;
+  case 'U':
+    gencast(fout,n,Gusmall);
+    return 1;
   case 'L':
   case 'P':
     gencast(fout,n,Gsmall);
@@ -147,10 +160,10 @@ int genarg(int nerr, FILE *fout, char c, int n, PPproto pr)
   case '&':
     if (tree[n].f!=Frefarg && pr == PPdefault)
       die(nerr,"Missing & for reference");
+  case '*': /* Fall through */
     fprintf(fout,"&");
     gencode(fout,n);
     return 1;
-  case '*':
   case 'W':
     gencode(fout,n);
     return 1;
@@ -297,10 +310,10 @@ void genuserfunc(FILE *fout, int n, int nf)
     firstarg=1;
     if (j<nb && arg[j]>=0 && arg[j]!=GNOARG)
       gencast(fout,arg[j],t);
-    else if (v->initval<0)
-      die(n,"a mandatory argument is missing");
     else if (v->flag&(1<<Cdefmarker))
       gencodenoarg(fout,t,n);
+    else if (v->initval<0)
+      die(n,"a mandatory argument is missing");
     else
       gencast(fout,v->initval,t);
   }
@@ -412,6 +425,8 @@ void gendecvar(FILE *fout, context *fc, int nerr)
 {
   int oldt=-1;
   int idx;
+  if (fc->s.n>0)
+    genlineno(fout, fc->c->node);
   for(idx=0;idx<fc->s.n;idx++)
   {
     ctxvar *v=fc->c+idx;
@@ -426,6 +441,8 @@ void gendecvar(FILE *fout, context *fc, int nerr)
     {
       gendecvarend(fout,oldt);
       genindent(fout);
+      if (idx<fc->s.n-1)
+        genlineno(fout, (v+1)->node);
       if (genfuncbydesc1(fout,v->node,FC_decl_base,nerr))
         die(nerr,"type `%s' not suitable for a variable",GPname(t));
       fprintf(fout, " ");
@@ -470,6 +487,9 @@ genprotoargs(FILE *fout, int nf, int firstarg, int nb, int *name)
     case 'n':
       fprintf(fout,"long");
       break;
+    case 'U':
+      fprintf(fout,"ulong");
+      break;
     case 'f':
       fprintf(fout,"long *");
       break;
@@ -486,7 +506,7 @@ genprotoargs(FILE *fout, int nf, int firstarg, int nb, int *name)
       break;
     case 'r':
     case 's':
-      fprintf(fout,"char *");
+      fprintf(fout,"const char *");
       break;
     default:
       die(err_func,"prototype letter `%c' not known",c);
@@ -544,7 +564,6 @@ void genprototype(FILE *fout, int nf, int kb)
         printnode(fout,v->initval);
         fprintf(fout,"*/");
       }
-
     }
   }
   s_ctx.n=savc;
@@ -711,6 +730,7 @@ void gendeffunc(FILE *fout, int n)
     die(n,"Internal error in gendeffunc : func %s not found",name);
   gp=lfunc+currfunc;
   gencomment(fout,funcid,0);
+  genlineno(fout, funcid);
   genprototype(fout,currfunc,1);
   t=functype(*gp);
   if (t!=Ggen && t!=Gsmall)
@@ -748,8 +768,7 @@ void gendefblock(FILE *fout, int n)
   m=tree[n].m;
   if(!(m&(1<<Mbrace)))
   {
-    genindent(fout);
-    fprintf(fout,"{\n");
+    genindentline(fout,"{\n");
     indent++;
   }
   gendecvar(fout,fc,n);
@@ -759,10 +778,7 @@ void gendefblock(FILE *fout, int n)
   if(!(m&(1<<Mbrace)))
   {
     indent--;
-    genindent(fout);
-    fprintf(fout,"}");
-    if(!(m&(1<<Muntil)))
-      fprintf(fout,"\n");
+    genendbrace(fout,n);
   }
   s_ctx.n=savc;
 }
