@@ -1,5 +1,5 @@
 /*
-Copyright (C) 2000-2005  The PARI group.
+Copyright (C) 2000-2013  The PARI group.
 
 This file is part of the GP2C package.
 
@@ -10,7 +10,7 @@ ANY WARRANTY WHATSOEVER.
 
 Check the License for details. You should have received a copy of it, along
 with the package; see the file 'COPYING'. If not, write to the Free Software
-Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. */
+Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.*/
 
 /*Patch for list of gp functions. Must be include after header.h.*/
 typedef enum {GPuser_private,
@@ -24,9 +24,10 @@ typedef enum {GPuser_private,
               GPglobal,GPlocal,
               GPif,
               GPprod,GPsum,GPsumdiv,GPprodeuler,
-              GPuntil,GPwhile,
+              GPuntil,GPwhile,GPwhilenull,
               GPmakevec,GPmakemat,
               GPvector,GPvectorsmall,GPvectorv,GPmatrix,
+              GPiferr,
               GPinitfunc,GPcopyarg,
               GPclosure
 } GPspec_private;
@@ -41,14 +42,18 @@ struct patchfunc_s
 } patchfunc[]={
 /*Override previous definition in pari/src/init.c*/
 {"O",NULL,Ggen,0,GPgrandO},
+{"addhelp",NULL,Gnotype,(1<<Msemicolon),GPaddhelp},
 {"break",NULL,Gvoid,0,GPbreak},
 {"next",NULL,Gvoid,0,GPnext},
 {"for","GGIG",Gnotype,(1<<Msemicolon),GPfor},
 {"fordiv",NULL,Gnotype,(1<<Msemicolon),GPfordiv},
 {"forstep","GIGGGG",Gnotype,(1<<Msemicolon),GPforstep},
-{"forvec",NULL,Gnotype,(1<<Msemicolon),GPforvec},
-{"forprime",NULL,Gnotype,(1<<Msemicolon),GPforprime},
+{"global",NULL,Gnotype,(1<<Msemicolon),GPglobal},
 {"if","GII",Gnotype,(1<<Msemicolon),GPif},
+{"iferr","IEII",Gnotype,(1<<Msemicolon),GPiferr},
+{"install",NULL,Gnotype,(1<<Msemicolon),GPinstall},
+{"local",NULL,Gnotype,(1<<Msemicolon),GPlocal},
+{"my",NULL,Gnotype,(1<<Msemicolon),GPlocal},
 {"prod","NULL",Gnotype,(1<<Msemicolon),GPprod},
 {"prodeuler",NULL,Gnotype,(1<<Msemicolon),GPprodeuler},
 {"return",NULL,Gnotype,(1<<Mterm),GPreturn},
@@ -60,18 +65,14 @@ struct patchfunc_s
 {"vectorv",NULL,Gnotype,(1<<Msemicolon),GPvectorv},
 {"matrix",NULL,Gnotype,(1<<Msemicolon),GPmatrix},
 {"while","EI",Gnotype,(1<<Msemicolon),GPwhile},
-{"_doloop","I",Gnotype,(1<<Msemicolon),GPdoloop},
 /* New definition*/
-{"global",NULL,Gnotype,(1<<Msemicolon),GPglobal},
-{"local",NULL,Gnotype,(1<<Msemicolon),GPlocal},
-{"my",NULL,Gnotype,(1<<Msemicolon),GPlocal},
-{"_makevec","*",Gnotype,(1<<Msemicolon),GPmakevec},
+{"_closure",NULL,Gnotype,0,GPclosure},
+{"_copyarg","",Gvoid,(1<<Msemicolon),GPcopyarg},
+{"_doloop","I",Gnotype,(1<<Msemicolon),GPdoloop},
+{"_forstepvec","GIGGGGGG",Gnotype,(1<<Msemicolon),GPforstepvec},
 {"_gerepilemany",NULL,Gvoid,(1<<Msemicolon),GPgpmany},
 {"_initfunc","",Gvoid,(1<<Msemicolon),GPinitfunc},
-{"_copyarg","",Gvoid,(1<<Msemicolon),GPcopyarg},
 {"_label",NULL,Gvoid,0,GPlabel},
-{"forstepvec","GIGGGG",Gnotype,(1<<Msemicolon),GPforstepvec},
-{"install",NULL,Gnotype,(1<<Msemicolon),GPinstall},
-{"addhelp",NULL,Gnotype,(1<<Msemicolon),GPaddhelp},
-{"_closure",NULL,Gnotype,0,GPclosure},
+{"_makevec","*",Gnotype,(1<<Msemicolon),GPmakevec},
+{"_whilenull","EI",Gnotype,(1<<Msemicolon),GPwhilenull},
 {NULL,NULL,0,0,0}};
